from clease.montecarlo.constraints import MCConstraint


class CollectiveVariableConstraint(MCConstraint):
    """
    Constraint that ensures that the collective variable defined by the getter
    stays within certain bounds

    Parameters:

    xmin: float
        Minimum value for the collective variable

    xmax: float
        Maximum value for the collective variable

    getter: MCObserver
        MCObsrever that support peak keyword that returns the collective
        variable after the proposed move
    """

    def __init__(self, xmin=0.0, xmax=1.0, getter=None):
        self.xmin = xmin
        self.xmax = xmax
        self.getter = getter

    def __call__(self, system_change):
        x = self.getter(system_change, peak=True)
        return x >= self.xmin and x < self.xmax



import numpy as np
from ase import Atoms
from clease.tools import equivalent_deco, nested_array2list, list2str, cname_lt
from .cluster_fingerprint import ClusterFingerprint

__all__ = ('Cluster',)


# pylint: disable=too-many-instance-attributes
class Cluster:

    def __init__(self,
                 name='c0',
                 size=0,
                 diameter=0.0,
                 fingerprint=None,
                 ref_indx=0,
                 indices=(),
                 equiv_sites=(),
                 trans_symm_group=0):
        self.name = name
        self.size = size
        self.diameter = diameter
        self.fp = fingerprint
        self.ref_indx = ref_indx
        self.indices = indices
        self.equiv_sites = equiv_sites
        self.group = trans_symm_group
        self.info = {}

    def __lt__(self, other):
        """Comparison operator."""
        fp_lt = self.fp < other.fp
        if fp_lt:
            assert cname_lt(self.name, other.name) is True
        else:
            assert cname_lt(self.name, other.name) is False
        return fp_lt

    def __eq__(self, other):
        """Equality operator."""
        fp_equal = self.fp == other.fp
        if fp_equal:
            assert self.name == other.name
        return fp_equal

    def __ne__(self, other):
        return not self.__eq__(other)

    def equiv_deco(self, deco):
        return equivalent_deco(deco, self.equiv_sites)

    def todict(self):
        """Return a dictionary representation."""
        return {
            'indices': self.indices,
            'size': self.size,
            'symm': self.group,
            'diameter': self.diameter,
            'name': self.name,
            'fingerprint': self.fp.todict(),
            'ref_indx': self.ref_indx,
            'equiv_sites': self.equiv_sites,
            'info': self.info
        }

    def from_dict(self, data):
        self.indices = data['indices']
        self.size = data['size']
        self.group = data['symm']
        self.diameter = data['diameter']
        self.name = data['name']

        if isinstance(data['fingerprint'], ClusterFingerprint):
            self.fp = data['fingerprint']
        elif isinstance(data['fingerprint'], dict):
            self.fp = ClusterFingerprint.load(data['fingerprint'])
        else:
            raise ValueError('Fingerprint has to be either instance of '
                             'ClusterFingerprint or a dictionary')

        self.ref_indx = data['ref_indx']
        self.equiv_sites = data['equiv_sites']
        self.info = data['info']

    @staticmethod
    def load(data):
        cluster = Cluster(None, None, None, None, None, None, None, None)
        fp = ClusterFingerprint.load(data['fingerprint'])
        data['fingerprint'] = fp
        data['indices'] = nested_array2list(data['indices'])
        data['equiv_sites'] = nested_array2list(data['equiv_sites'])
        cluster.from_dict(data)
        return cluster

    def is_subcluster(self, other):
        """Check if the passed cluster is a subcluster of the current."""
        if len(self.indices) == 0:
            return True

        if len(self.indices[0]) >= len(other.indices[0]):
            return False

        return any(set(s1).issubset(s2) for s1 in self.indices for s2 in other.indices)

    def get_figure(self, generator) -> Atoms:
        """Get figure from a ClusterGenerator object"""
        if len(self.indices[0][0]) != 4:
            raise RuntimeError("This method requires that the cluster is "
                               "described based on its 4-vector and not index "
                               "in the ASE atoms")

        positions = np.array([generator.cartesian(x) for x in self.indices[0]])
        positions -= np.mean(positions, axis=0)
        sublat_symb = {at.tag: at.symbol for at in generator.prim}
        symbols = [sublat_symb[x[3]] for x in self.indices[0]]
        return Atoms(symbols, positions=positions)

    def __str__(self):
        str_rep = f"Name: {self.name}\n"
        str_rep += f"Diameter: {self.diameter}\n"
        str_rep += f"Size: {self.size}\n"
        str_rep += f"Ref. indx: {self.ref_indx}\n"
        str_rep += f"Trans. symm group: {self.group}\n"
        str_rep += f"Indices: {self.indices}\n"
        str_rep += f"Equiv. sites: {self.equiv_sites}\n"
        str_rep += f"Fingerprint: {self.fp}\n"
        str_rep += f"Information: {self.info}\n"
        return str_rep

    def get_figure_key(self, figure):
        """Return a key representation of the figure."""
        return list2str(self._order_equiv_sites(figure))

    def _order_equiv_sites(self, figure) -> list:
        """Sort equivalent sites."""
        figure_cpy = list(figure)
        for eq_group in self.equiv_sites:
            equiv_indices = [figure[i] for i in eq_group]
            equiv_indices.sort()
            for count, i in enumerate(eq_group):
                figure_cpy[i] = equiv_indices[count]
        return figure_cpy

    @property
    def num_fig_occurences(self):
        """Number of currences for each figures."""
        occ_count = {}
        for figure in self.indices:
            key = self.get_figure_key(figure)
            current_num = occ_count.get(key, 0)
            occ_count[key] = current_num + 1
        return occ_count

    def corresponding_figure(self, ref_indx, target_figure, trans_matrix):
        """Find figures that correspond to another reference index.

        Parameters:

        ref_indx: int
            reference index

        taget_figres: list of indices
            list of atomic indices that constitute the original figure before
            translating

        trans_matrix: list of dicts
            translation matrix
        """
        for figure in self.indices:
            translated_figure = [trans_matrix[ref_indx][x] for x in figure]
            translated_figure = self._order_equiv_sites(translated_figure)
            if translated_figure == target_figure:
                return self._order_equiv_sites(figure)

        raise RuntimeError(f"There are no matching figure for ref_indx: "
                           f"{ref_indx} and figure: {target_figure}!")

    def get_all_figure_keys(self):
        return [self.get_figure_key(fig) for fig in self.indices]



import os
import logging
from numpy.random import shuffle, choice
import numpy as np
from clease.tools import (aic, aicc, bic)
from .regression import LinearRegression

logger = logging.getLogger(__name__)

workers = None

__all__ = ('SaturatedPopulationError', 'GAFit')


class SaturatedPopulationError(Exception):
    pass


# pylint: disable=too-many-instance-attributes
class GAFit:
    """
    Genetic Algorithm for selecting relevant clusters.

    Parameters:

    cf_matrix: np.ndarray
        Design matrix of the linear regression (nxm) where n is the number of
        data points and m is the number of features

    e_dft: list
        Array of length n with DFT energies

    elitism: int
        Number of best structures that will be passed unaltered on to the next
        generation

    fname: str
        File name used to backup the population. If this file exists, the next
        run will load the population from the file and start from there.
        Another file named 'fname'_cf_names.txt is created to store the
        names of selected clusters.

    num_individuals: int or str
        Integer with the number of inidivuals or it is equal to "auto",
        in which case 10 times the number of candidate clusters is used

    max_num_in_init_pool: int
        If given the maximum clusters included in the initial population
        is given by this number. If max_num_in_init_pool=150, then
        solution with maximum 150 will be present in the initial pool.

    cost_func: str
        Use the inverse as fitness measure.
        Possible cost functions:
        bic - Bayes Information Criterion
        aic - Afaike Information Criterion
        aicc - Modified Afaikes Information Criterion
        (tend to avoid overfitting better than aic)
    """

    def __init__(self,
                 cf_matrix,
                 e_dft,
                 mutation_prob=0.001,
                 elitism=1,
                 fname="ga_fit.csv",
                 num_individuals="auto",
                 max_num_in_init_pool=None,
                 cost_func="aicc"):
        allowed_cost_funcs = {"bic": bic, "aic": aic, "aicc": aicc}

        if cost_func not in allowed_cost_funcs:
            raise ValueError(f"Cost func has to be one of {allowed_cost_funcs.keys()}")

        self.cost_func = allowed_cost_funcs[cost_func]

        # Read required attributes from evaluate
        self.cf_matrix = cf_matrix
        self.e_dft = e_dft
        self.fname = fname
        self.fname_cf_names = fname.rpartition(".")[0] + "_cf_names.txt"

        if num_individuals == "auto":
            self.pop_size = 10 * self.cf_matrix.shape[1]
        else:
            self.pop_size = int(num_individuals)

        # Make sure that the population size is an even number
        if self.pop_size % 2 == 1:
            self.pop_size += 1
        self.num_genes = self.cf_matrix.shape[1]
        self.individuals = self._initialize_individuals(max_num_in_init_pool)
        self.fitness = np.zeros(len(self.individuals))
        self.regression = LinearRegression()
        self.elitism = elitism
        self.mutation_prob = mutation_prob
        self.statistics = {"best_score": [], "worst_score": []}
        self.evaluate_fitness()
        self.check_valid()

    def _initialize_individuals(self, max_num):
        """Initialize a random population."""
        individuals = []
        if os.path.exists(str(self.fname)):
            individuals = self._init_from_file()
        else:
            max_num = max_num or self.num_genes
            indices = list(range(self.num_genes))
            num_non_zero = np.array(list(range(0, self.pop_size)))
            num_non_zero %= max_num
            num_non_zero[num_non_zero < 3] = 3
            for i in range(self.pop_size):
                shuffle(indices)
                individual = np.zeros(self.num_genes, dtype=np.uint8)
                indx = indices[:num_non_zero[i]]
                individual[np.array(indx)] = 1
                individuals.append(self.make_valid(individual))
        return individuals

    def _init_from_file(self):
        """Initialize the population from file."""
        logger.info("Initializing population from %s", self.fname)
        individuals = []
        with open(self.fname, 'r') as infile:
            for line in infile:
                individual = np.zeros(self.num_genes, dtype=np.uint8)
                indices = np.array([int(x.strip()) for x in line.split(",")])
                individual[indices] = 1
                individuals.append(individual)
        return individuals

    def get_eci(self, individual):
        """Calculate the LOOCV for the current individual."""
        X = self.design_matrix(individual)
        y = self.e_dft
        return self.regression.fit(X, y)

    def design_matrix(self, individual):
        """Return the corresponding design matrix."""
        return self.cf_matrix[:, individual == 1]

    def fit_individual(self, individual):
        coeff = self.get_eci(individual)
        X = self.design_matrix(individual)
        e_pred = X.dot(coeff)
        delta_e = self.e_dft - e_pred

        info_measure = None
        n_selected = np.sum(individual)
        mse = np.sum(delta_e**2) / self.num_data
        info_measure = self.cost_func(mse, n_selected, self.num_data)
        return coeff, -info_measure

    def evaluate_fitness(self):
        """Evaluate fitness of all species."""
        for i, ind in enumerate(self.individuals):
            _, fit = self.fit_individual(ind)
            self.fitness[i] = fit

    @staticmethod
    def flip_one_mutation(individual):
        """Apply mutation where one bit flips."""
        indx_sel = list(np.argwhere(individual.T == 1).T[0])
        ns = list(np.argwhere(individual.T == 0).T[0])

        assert len(indx_sel) != 0 or len(ns) != 0
        if len(ns) == 0:
            ns = indx_sel
        elif len(indx_sel) == 0:
            indx_sel = ns

        # Flip included or not included cluster with equal
        # probability
        if np.random.rand() < 0.5:
            indx = choice(indx_sel)
        else:
            indx = choice(ns)
        individual[indx] = (individual[indx] + 1) % 2
        return individual

    @staticmethod
    def make_valid(individual):
        """Make sure that there is at least two active ECIs."""
        if np.sum(individual) < 2:
            while np.sum(individual) < 2:
                indx = np.random.randint(low=0, high=len(individual))
                individual[indx] = 1
        return individual

    def create_new_generation(self):
        """Create a new generation."""
        new_generation = []
        srt_indx = np.argsort(self.fitness)[::-1]

        assert self.fitness[srt_indx[0]] >= self.fitness[srt_indx[1]]

        # Pass the fittest to the next generation
        num_transfered = 0
        counter = 0
        while num_transfered < self.elitism and counter < len(srt_indx):
            indx = srt_indx[counter]

            individual = self.individuals[indx].copy()

            # Transfer the best
            new_generation.append(individual)

            # Transfer the best individual with a mutation
            new_ind = self.flip_one_mutation(individual.copy())
            new_ind = self.make_valid(new_ind)
            while self._is_in_population(new_ind, new_generation):
                new_ind = self.flip_one_mutation(individual.copy())
                new_ind = self.make_valid(new_ind)

            new_generation.append(new_ind)
            num_transfered += 1
            counter += 1

        if counter >= len(srt_indx):
            self.save_population()
            raise SaturatedPopulationError("The entire population has been saturated!")

        only_positive = self.fitness - np.min(self.fitness)
        cumulative_sum = np.cumsum(only_positive)
        cumulative_sum /= cumulative_sum[-1]
        num_inserted = len(new_generation)

        max_attempts = 100 * self.pop_size
        # Create new generation by mergin existing
        for i in range(num_inserted, int(self.pop_size / 2) + 1):
            rand_num = np.random.rand()
            p1 = np.argmax(cumulative_sum > rand_num)
            p2 = p1
            while p2 == p1:
                rand_num = np.random.rand()
                p2 = np.argmax(cumulative_sum > rand_num)

            new_individual = self.individuals[p1].copy()
            new_individual2 = self.individuals[p2].copy()

            mask = np.random.randint(0, high=2, size=len(new_individual))
            mask = mask.astype(np.uint8)
            new_individual[mask] = self.individuals[p2][mask]
            new_individual2[mask] = self.individuals[p1][mask]
            new_individual = self.make_valid(new_individual)
            new_individual2 = self.make_valid(new_individual2)

            # Check if there are any equal individuals in
            # the population
            counter = 0
            while (self._is_in_population(new_individual, new_generation) and
                   counter < max_attempts):
                new_individual = self.flip_one_mutation(new_individual)
                new_individual = self.make_valid(new_individual)
                counter += 1

            if counter >= max_attempts:
                self.save_population()
                raise SaturatedPopulationError("Popluation is saturated!")
            new_generation.append(new_individual)

            counter = 0
            while (self._is_in_population(new_individual2, new_generation) and
                   counter < max_attempts):
                new_individual2 = self.flip_one_mutation(new_individual2)
                new_individual2 = self.make_valid(new_individual2)
                counter += 1

            if counter >= max_attempts:
                self.save_population()
                raise SaturatedPopulationError("Popluation is saturated!")

            new_generation.append(new_individual2)

        if len(new_generation) != len(self.individuals):
            raise RuntimeError(f"Size of generation changed! Original size: "
                               f"{len(self.individuals)}. New size: "
                               f"{len(new_generation)}")
        self.individuals = new_generation

    @staticmethod
    def _is_in_population(ind, pop):
        """Check if the individual is already in the population."""
        return any(np.all(ind == x) for x in pop)

    def mutate(self):
        """Introduce mutations."""
        avg_f = np.mean(np.abs(self.fitness))
        best_indx = np.argmax(self.fitness)
        for i in range(len(self.individuals)):
            if i == best_indx:
                # Do not mutate the best individual
                continue

            mut_prob = self.mutation_prob

            # Increase the probability of introducing mutations
            # to the least fit individuals
            if abs(self.fitness[i]) > avg_f:
                mut_prob *= abs(self.fitness[i]) / avg_f

            if mut_prob > 1.0:
                mut_prob = 1.0

            ind = self.individuals[i].copy()
            mutated = False
            assert mut_prob >= 0.0
            if np.random.rand() < mut_prob:
                ind = self.flip_one_mutation(ind)
                mutated = True

            if mutated:
                self.individuals[i] = self.make_valid(ind)
                _, fit = self.fit_individual(self.individuals[i])
                self.fitness[i] = fit

    def population_diversity(self):
        """Check the diversity of the population."""
        std = np.std(self.individuals)
        return np.mean(std)

    @property
    def best_individual(self):
        best_indx = np.argmax(self.fitness)
        individual = self.individuals[best_indx]
        return individual

    @property
    def num_data(self):
        return self.cf_matrix.shape[0]

    @property
    def best_individual_indx(self):
        best_indx = np.argmax(self.fitness)
        return best_indx

    def index_of_selected_clusters(self, individual):
        """Return the indices of the selected clusters

        Parameters:

        individual: int
            Index of the individual
        """
        return list(np.nonzero(self.individuals[individual])[0])

    def save_population(self):
        # Save population
        self.check_valid()
        with open(self.fname, 'w') as out:
            for i in range(len(self.individuals)):
                out.write(",".join(str(x) for x in self.index_of_selected_clusters(i)))
                out.write("\n")
        logger.info("Population written to %s.", self.fname)

    def plot_evolution(self):
        """Create a plot of the evolution."""
        from matplotlib import pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(self.statistics["best_score"], label="best")
        ax.plot(self.statistics["worst_score"], label="worst")
        ax.set_xlabel("Generation")
        ax.set_ylabel("Score")
        plt.show()

    def run(self, gen_without_change=100, min_change=0.01, save_interval=100):
        """Run the genetic algorithm.

        Return a list consisting of the names of selected clusters at the end
        of the run.

        Parameters:

        gen_without_change: int
            Terminate if gen_without_change are created without sufficient
            improvement

        min_change: float
            Changes a larger than this value is considered "sufficient"
            improvement

        save_interval: int
            Rate at which all the populations are backed up in a file
        """
        num_gen_without_change = 0
        current_best = 0.0
        gen = 0
        while True:
            self.evaluate_fitness()

            best_indx = np.argmax(self.fitness)

            num_eci = np.sum(self.individuals[best_indx])
            diversity = self.population_diversity()
            self.statistics["best_score"].append(np.max(self.fitness))
            self.statistics["worst_score"].append(np.min(self.fitness))

            best3 = np.abs(np.sort(self.fitness)[::-1][:3])
            loocv_msg = ""

            logger.info(("Generation: %s. Top 3 scores %.2e (-)%.2e (-)%.2e."
                         " Num ECI: %d. Pop. div: %.2f. %s"), gen, best3[0], best3[0] - best3[1],
                        best3[0] - best3[2], num_eci, diversity, loocv_msg)
            self.mutate()
            self.create_new_generation()
            if abs(current_best - self.fitness[best_indx]) > min_change:
                num_gen_without_change = 0
            else:
                num_gen_without_change += 1
            current_best = self.fitness[best_indx]

            if gen % save_interval == 0:
                self.save_population()

            if num_gen_without_change >= gen_without_change:
                logger.info('Reached %d generations without sufficient improvement.',
                            gen_without_change)
                break
            gen += 1

        self.save_population()
        return self.best_individual

    def check_valid(self):
        """Check that the current population is valid."""
        for ind in self.individuals:
            valid = self.make_valid(ind.copy())
            if np.any(valid != ind):
                raise ValueError("Individual violate constraints!")



from kivy.uix.screenmanager import Screen
from kivy.properties import StringProperty
from clease.gui.mc_main_page import MCMainPage
from clease.gui.canonical_mc_page import CanonicalMCPage
from kivy.app import App
from threading import Thread
from clease.gui.constants import SCREEN_TRANSLATIONS


class MCHeader(Screen):
    mc_type = StringProperty('Main')

    def change_screen(self, txt):
        all_mc_screens = self.ids.pageSpinner.values

        cur_screen = self.ids.sm.current
        cur_key = None
        for k, v in SCREEN_TRANSLATIONS.items():
            if v == cur_screen:
                cur_key = k
                break

        i_cur = all_mc_screens.index(cur_key)
        i_next = all_mc_screens.index(txt)

        if i_next > i_cur:
            self.ids.sm.transition.direction = 'left'
        else:
            self.ids.sm.transition.direction = 'right'
        self.ids.sm.current = SCREEN_TRANSLATIONS[txt]

    @property
    def mc_cell_size(self):
        page = self.ids.sm.get_screen('MCMainPage')
        return int(page.ids.sizeInput.text)



from typing import Dict
from ase import Atoms
from clease.corr_func import CorrFunction
from clease.settings import ClusterExpansionSettings
from clease.calculator import Clease
from clease.tools import wrap_and_sort_by_position


def attach_calculator(settings: ClusterExpansionSettings,
                      atoms: Atoms,
                      eci: Dict[str, float] = None) -> Atoms:
    """Utility function for an efficient initialization of large cells.

    :param settings: ClusterExpansionSettings object (e.g., CEBulk, CECyrstal)
    :param eci: Dictionary containing cluster names and their ECI values
    :param atoms: ASE Atoms object.
    """
    if eci is None:
        eci = {}
    cf_names = list(eci.keys())
    init_cf = CorrFunction(settings).get_cf_by_names(settings.atoms, cf_names)

    template = settings.template_atoms.get_template_matching_atoms(atoms=atoms)
    settings.set_active_template(atoms=template)

    wrapped = wrap_and_sort_by_position(atoms.copy())
    atoms_with_calc = settings.atoms.copy()

    calc = Clease(settings, eci=eci, init_cf=init_cf)
    atoms_with_calc.calc = calc
    atoms_with_calc.symbols = wrapped.symbols
    return atoms_with_calc


def get_ce_energy(settings: ClusterExpansionSettings, atoms: Atoms, eci: Dict[str, float]) -> float:
    """Get energy of the ASE Atoms object based on given ECI values.

    :param settings: ClusterExpansionSettings object (e.g., CEBulk, CECyrstal)
    :param atoms: ASE Atoms object representing the considered atomic
        arrangement
    :param eci: ECI values to be used to calculate the energy
    """
    # temp atoms to avoid making unexpected changes to passed atoms
    temp_atoms = atoms.copy()
    calc = Clease(settings, eci=eci)
    temp_atoms.calc = calc
    energy = temp_atoms.get_potential_energy()
    return energy



from clease.gui.legend_item import LegendItem


class Legend(object):

    def __init__(self, layout):
        self.layout = layout

    def setup(self, new_items, num_rows=1, num_cols=1):
        """
        Initialise the legend

        Parameters:

        items: List of dict
            List of dictionaries for each item. The dictionary
            constain {'text': legend_text, 'color': rgb color}

        num_rows: int
            Number of rows

        num_cols: int
            Number of columns
        """
        self.layout.clear_widgets()
        if num_rows * num_cols < len(new_items):
            raise ValueError('num_rows*num_cols < len(items)!')

        counter = 0
        for r in range(num_rows):
            for c in range(num_cols):
                item = LegendItem(icon_color=new_items[counter]['color'],
                                  text=new_items[counter]['text'],
                                  icon_size=[20, 20])
                self.layout.add_widget(item)
                counter += 1
                if counter >= len(new_items):
                    return



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure
from clease import Evaluate, ConvexHull


def plot_fit(evaluate: Evaluate, plot_args: dict = None) -> Figure:
    """
    Figure object calculated (DFT) and predicted energies.
    If the plot_args dictionary contains keys,
    return  figure object to relate plot_args  keys

    :param evaluate: Use the evaluate object
    to define the plot argument.
    :param plot_args: plot_args dictionary contains:
        - "xlabel": x-axis label
        - "ylabel": y-axis label
        - "title": title of plot

    :return: Figure instance of plot
    """
    if plot_args is None:
        plot_args = {}
    X = evaluate.get_energy_predict()
    Y = evaluate.e_dft
    xlabel = plot_args.get("xlabel", "E_DFT (eV/atom)")
    ylabel = plot_args.get("ylabel", "E_CE (eV/atom)")
    title = plot_args.get("title", f"Fit using {len(evaluate.e_dft)} data points.")

    # rmin, rmax set the plot range of x, y coordinate
    if np.size(X) and np.size(Y) != 0:
        e_range = max(np.append(X, Y)) - min(np.append(X, Y))
        rmin = min(np.append(X, Y)) - 0.05 * e_range
        rmax = max(np.append(X, Y)) + 0.05 * e_range
    else:
        rmin = -10
        rmax = 10
    linear_fit = np.arange(rmin - 10, rmax + 10, 1)
    cv_name = evaluate.scoring_scheme.lower()

    if cv_name == 'k-fold':
        # Figure out the k
        nsplits = evaluate.nsplits
        cv_name = f'{nsplits}-fold CV'

    cv = evaluate.get_cv_score()
    rmse = evaluate.rmse() * 1000

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.set_title(title)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)
    ax.axis([rmin, rmax, rmin, rmax])
    ax.text(0.95,
            0.01,
            cv_name + f" = {cv:.3f} meV/atom\n"
            f"RMSE = {rmse:.3f} meV/atom",
            verticalalignment='bottom',
            horizontalalignment='right',
            transform=ax.transAxes,
            fontsize=12)
    ax.plot(linear_fit, linear_fit, 'r')
    ax.plot(X, Y, 'bo', mfc='none')

    return fig


def plot_fit_residual(evaluate: Evaluate, plot_args: dict = None) -> Figure:
    """
    Figure object subtracted (DFT) and predicted energies.
    If the plot_args dictionary contains keys,
    return  figure object to relate plot_args  keys

    :param evaluate: Use the evaluate object
    to define the plot argument.
    :param plot_args: plot_args dictionary contains:
        - "xlabel": x-axis label
        - "ylabel": y-axis label
        - "title": title of plot

    :return: Figure instance of plot
    """
    if plot_args is None:
        plot_args = {}
    X = evaluate.e_dft
    Y = evaluate.subtract_predict_dft()
    xlabel = plot_args.get("xlabel", "#OCC")
    ylabel = plot_args.get("ylabel", r"$E_{DFT} - E_{pred}$ (meV/atom)")
    title = plot_args.get("title", "Residual (v)")

    gridspec_kw = {"wspace": 0.0, "width_ratios": [5, 1]}
    fig, ax = plt.subplots(ncols=2, sharey=True, gridspec_kw=gridspec_kw)
    ax[0].set_title(title)
    ax[0].set_ylabel(ylabel)
    ax[0].axhline(0, ls="--")
    ax[0].plot(X, Y, "v", mfc="none")

    hist, bin_edges = np.histogram(Y, bins=30)
    h = bin_edges[1] - bin_edges[0]
    ax[1].barh(bin_edges[:-1], hist, height=h, color="#bdbdbd")
    ax[1].set_xlabel(xlabel)
    ax[1].spines["right"].set_visible(False)
    ax[1].spines["top"].set_visible(False)

    return fig


def plot_eci(evaluate: Evaluate, plot_args: dict = None) -> Figure:
    """
    Figure object of ECI value according to cluster diameter
    If the plot_args dictionary contains keys,
    return  figure object to relate plot_args  keys

    :param evaluate: Use the evaluate object
    to define the plot argument.
    :param plot_args: plot_args dictionary contains:
        - "xlabel": x-axis label
        - "ylabel": y-axis label
        - "title": title of plot
        - "sizes": list of int to include n-body cluster in plot

    :return: Figure instance of plot
    """
    if plot_args is None:
        plot_args = {}
    # eci_by_size dictionary contains eci, name, distance
    eci_by_size = evaluate.get_eci_by_size()
    xlabel = plot_args.get("xlabel", "Cluster diameter ($n^{th}$ nearest neighbor)")
    ylabel = plot_args.get("ylabel", "ECI (eV/atom)")
    title = plot_args.get("title", "Plot ECI")
    sizes = plot_args.get("sizes", list(eci_by_size.keys()))

    markers = ["o", "v", "x", "D", "^", "h", "s", "p"]
    lines = []

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.axhline(0.0, ls="--", color="grey")
    ax.xaxis.set_major_locator(plt.MultipleLocator(1))
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    for size in sizes:
        data = eci_by_size[size]
        # Add 1, as NN starts from 1 and not 0
        X = np.array(data["distance"]) + 1
        Y = data["eci"]
        mrk = markers[size % len(markers)]
        line = ax.plot(X, Y, label=f"{size}-body", marker=mrk, mfc="none", ls="", markersize=8)
        lines.append(line[0])
    ax.legend()
    return fig


def plot_cv(evaluate: Evaluate, plot_args: dict = None) -> Figure:
    """
    Figure object of CV values according to alpha values
    If the plot_args dictionary contains keys,
    return  figure object to relate plot_args  keys

    :param evaluate: Use the evaluate object
    to define the plot argument.
    :param plot_args: plot_args dictionary contains:
        - "xlabel": x-axis label
        - "ylabel": y-axis label
        - "title": title of plot

    :return: Figure instance of plot
    """
    if plot_args is None:
        plot_args = {}
    alpha_cv_data = evaluate.cv_scores
    xlabel = plot_args.get("xlabel", "alpha")
    ylabel = plot_args.get("ylabel", 'CV score (meV/atom)')
    title = plot_args.get("title", 'CV score vs. alpha')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.set_title(title)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)

    # this variable is for getting minimum cv
    min_cv = min(alpha_cv_data, key=lambda x: x['cv'])
    X = []
    Y = []
    for data in alpha_cv_data:
        X.append(data['alpha'])
        Y.append(data['cv'])
    ax.plot(X, Y)
    ax.text(0.65,
            0.01, f"min. CV score:\n"
            f"alpha = {min_cv['alpha']:.10f} \n"
            f"CV = {min_cv['cv'] * 1000.0:.3f}"
            f" meV/atom",
            verticalalignment='bottom',
            horizontalalignment='left',
            transform=ax.transAxes,
            fontsize=10)
    return fig


def plot_convex_hull(evaluate: Evaluate) -> Figure:

    e_pred = evaluate.get_energy_predict()

    cnv_hull = ConvexHull(evaluate.settings.db_name, select_cond=evaluate.select_cond)
    fig = cnv_hull.plot()  # Construct the figure object

    # `conc_per_frame` is the concentration with respect to the total number of atoms for
    # each frame
    conc_per_frame = evaluate.atomic_concentrations

    # `concs` is dictionary with the keys of species with value being the
    # concentrations among the frames
    # pylint: disable=protected-access
    concs = {key: [] for key in cnv_hull._unique_elem}
    for frame_conc in conc_per_frame:
        for key, value in concs.items():
            value.append(frame_conc.get(key, 0.0))

    form_en = [cnv_hull.get_formation_energy(c, e) for c, e in zip(conc_per_frame, e_pred.tolist())]
    cnv_hull.plot(fig=fig, concs=concs, energies=form_en, marker="x")

    fig.suptitle("Convex hull DFT (o), CE (x)")

    return fig



"""Class containing a manager for creating template atoms."""
from itertools import product
from typing import Iterator, List
from contextlib import contextmanager
import numpy as np
from numpy.random import shuffle
import ase
from ase.build import make_supercell
from clease import SkewnessFilter, EquivalentCellsFilter
from clease.template_filters import CellFilter, AtomsFilter
from clease.tools import all_integer_transform_matrices

__all__ = ('TemplateAtoms',)


class TemplateAtoms:

    def __init__(self, prim_cell, supercell_factor=27, size=None, skew_threshold=40, filters=()):
        if size is None and supercell_factor is None:
            raise TypeError("Either size or supercell_factor needs to be "
                            "specified.\n size: list or numpy array.\n "
                            "supercell_factor: int")

        self.supercell_factor = supercell_factor
        self._skew_threshold = skew_threshold
        self.size = size

        self.cell_filters = []
        self.atoms_filters = []

        self.add_cell_filter(SkewnessFilter(skew_threshold))

        for f in filters:
            if isinstance(f, CellFilter):
                self.add_cell_filter(f)
            elif isinstance(f, AtomsFilter):
                self.add_atoms_filter(f)

        if self.size is not None:
            self.supercell_factor = None
            check_valid_conversion_matrix(self.size)
        self.prim_cell = prim_cell

    def __eq__(self, other):
        return self.supercell_factor == other.supercell_factor and \
            self.skew_threshold == other.skew_threshold and \
            self.size == other.size and self.prim_cell == other.prim_cell

    @property
    def skew_threshold(self):
        return self._skew_threshold

    @skew_threshold.setter
    def skew_threshold(self, threshold):
        self._skew_threshold = threshold
        for f in self.cell_filters:
            if isinstance(f, SkewnessFilter):
                f.ratio = threshold

    def __str__(self):
        """Print a summary of the class."""
        msg = "=== TemplateAtoms ===\n"
        msg += f"Supercell factor: {self.supercell_factor}\n"
        msg += f"Size: {self.size}\n"
        msg += f"Skew threshold: {self.skew_threshold}\n"
        return msg

    def add_cell_filter(self, cell_filter):
        """Attach a new Cell filter."""
        if not isinstance(cell_filter, CellFilter):
            raise TypeError("filter has to be an instance of CellFilter")
        self.cell_filters.append(cell_filter)

    def add_atoms_filter(self, at_filter):
        """Attach a new Atoms filter."""
        if not isinstance(at_filter, AtomsFilter):
            raise TypeError("filter has to be an instance of AtomsFilter")
        self.atoms_filters.append(at_filter)

    def clear_filters(self):
        """Remove all filters."""
        self.cell_filters = []
        self.atoms_filters = []

    def remove_filter(self, f):
        """Remove one filter"""
        if isinstance(f, AtomsFilter):
            self.atoms_filters.remove(f)
        elif isinstance(f, CellFilter):
            self.cell_filters.remove(f)
        else:
            raise TypeError('Only AtomsFilters and CellFilters can be removed')

    def remove_filters(self, filters):
        """Remove a list of filters."""
        for f in filters:
            self.remove_filter(f)

    def is_valid(self, atoms=None, cell=None):
        """
        Check the validity of the template.

        Return `True` if templates are valid according to the attached filters.

        Parameters:

        atoms: Atoms object

        cell: unit cell vector
        """
        if atoms is None and cell is None:
            msg = "At least one of `atoms` or `cell` must be specified."
            raise ValueError(msg)

        if cell is None:
            cell = atoms.get_cell()
        cell_valid = all(f(cell) for f in self.cell_filters)

        if not cell_valid:
            return False

        atoms_valid = True
        if atoms is not None:
            atoms_valid = all(f(atoms) for f in self.atoms_filters)

        return cell_valid and atoms_valid

    def get_template_with_given_size(self, size):
        """Get the UID of the template with given size.

        Parameters:

        size: list of length 3
        """
        if not is_3x3_matrix(size):
            raise ValueError("Expect the size to be a list of 3x3 matrix")

        check_valid_conversion_matrix(size)

        template = make_supercell(self.prim_cell, size)

        if isinstance(size, np.ndarray):
            size = size.tolist()
        template.info['size'] = size

        if not self.is_valid(atoms=template):
            raise ValueError("Requested size violates the constraints!")
        return template

    def get_template_matching_atoms(self, atoms):
        """Get the UID for the template matching atoms.

        Parameters:

        atoms: Atoms object
            structure to compare its size against template atoms
        """

        size = self._get_conversion_matrix(atoms)
        assert is_3x3_matrix(size)

        prim_cell = self.prim_cell

        if isinstance(size, np.ndarray):
            size = size.tolist()

        template = make_supercell(prim_cell, size)
        template.info['size'] = size

        if not np.allclose(atoms.get_cell(), template.get_cell()):
            raise ValueError(f"Inconsistent cells! Passed atoms\n"
                             f"{atoms.get_cell()}\nGenerated template\n"
                             f"{template.get_cell()}")

        if not self.is_valid(atoms=template):
            raise ValueError("Requested template violates the constraints!")
        return template

    def _get_conversion_matrix(self, atoms):
        """Return the conversion matrix factor."""
        prim_cell = self.prim_cell

        small_cell = prim_cell.get_cell()
        inv_cell = np.linalg.inv(small_cell)

        large_cell = atoms.get_cell()
        size_factor = large_cell.dot(inv_cell)
        scale_int = size_factor.round(decimals=0).astype(int)
        if np.allclose(size_factor, scale_int):
            check_valid_conversion_matrix(scale_int)
            return scale_int.tolist()

        raise ValueError(f"The passed atoms object cannot be described by "
                         f"repeating of the unit cells. Scale factors found "
                         f"{size_factor}")

    def get_all_templates(self) -> List[ase.Atoms]:
        """Return a list with all templates."""
        return list(self.iterate_all_templates())

    def iterate_all_templates(self, max_per_size: int = None) -> Iterator[ase.Atoms]:
        """Get ass possible templates in an iterator.

        :param max_per_size: Maximum number of iterations per size.
            Optional. If None, then all sizes will be used.
        """
        if self.size is not None:
            yield self.get_template_with_given_size(self.size)
            return

        for size in range(1, self.supercell_factor):
            for ii, template in enumerate(self._iterate_templates_at_size(size)):
                # Check if we limit the number of tempaltes per size
                if max_per_size is not None and ii >= max_per_size:
                    break
                yield template

    def _iterate_templates_at_size(self, size: int) -> Iterator[ase.Atoms]:
        """Get all templates at a given size, i.e.
        a size at a given supercell factor
        """
        cells = []
        equiv_filter = EquivalentCellsFilter(cells)
        ucell = self.prim_cell.get_cell()

        matrices = all_integer_transform_matrices(size)

        for mat in matrices:
            cell = mat.dot(ucell)
            new_atoms_flag = False  # Did we create a new template?
            with self.filter_context(equiv_filter):
                if self.is_valid(cell=cell):
                    at = make_supercell(self.prim_cell, mat)
                    if self.is_valid(atoms=at):
                        at.info['size'] = mat.tolist()
                        cells.append(cell)
                        new_atoms_flag = True
            # Ensure we remove the filter again, so it does not affect the
            # global filters when processing these atoms outside
            # of this function
            if new_atoms_flag:
                yield at

    @contextmanager
    def filter_context(self, custom_filter):
        try:
            self.add_cell_filter(custom_filter)
            yield
        finally:
            self.remove_filter(custom_filter)

    def get_all_scaled_templates(self):
        """
        Return all templates that can be obtained by scaling the primitive cell
        """
        if self.size is not None:
            return [self.get_template_with_given_size(self.size)]
        cells = []
        templates = []
        equiv_filter = EquivalentCellsFilter(cells)
        self.add_cell_filter(equiv_filter)
        ucell = self.prim_cell.get_cell()
        for diag in product(range(1, self.supercell_factor + 1), repeat=3):
            if np.prod(diag) > self.supercell_factor:
                continue

            mat = np.diag(diag)
            cell = mat.dot(ucell)
            if self.is_valid(cell=cell):
                at = make_supercell(self.prim_cell, mat)
                if self.is_valid(atoms=at):
                    cells.append(cell)
                    at.info['size'] = mat.tolist()
                    templates.append(at)
        self.remove_filter(equiv_filter)
        return templates

    def weighted_random_template(self):
        """Select a random template atoms with a bias towards a cubic cell.

        The bias is towards cells that have similar values for x-, y- and
        z-dimension sizes.
        """
        if self.size is not None:
            return self.get_template_with_given_size(self.size)

        ucell = self.prim_cell.get_cell()
        max_attempts = 100000
        exp_value = self.supercell_factor**(1.0 / 3.0)
        for _ in range(max_attempts):
            diag = np.random.poisson(lam=exp_value, size=3)
            while np.prod(diag) > self.supercell_factor or np.prod(diag) == 0:
                diag = np.random.poisson(lam=exp_value, size=3)

            off_diags = np.random.poisson(lam=1, size=3)

            # Check that off diagonals are smaller or equal to diagonal
            if (off_diags[0] > diag[0] or off_diags[1] > diag[0] or off_diags[2] > diag[1]):
                continue

            matrix = np.zeros((3, 3), dtype=int)
            matrix[0, 0] = diag[0]
            matrix[1, 1] = diag[1]
            matrix[2, 2] = diag[2]
            matrix[0, 1] = off_diags[0]
            matrix[0, 2] = off_diags[1]
            matrix[1, 2] = off_diags[2]
            cell = matrix.dot(ucell)

            if self.is_valid(cell=cell):
                atoms = make_supercell(self.prim_cell, matrix)
                if self.is_valid(atoms=atoms):
                    atoms.info['size'] = matrix.tolist()
                    return atoms

        raise RuntimeError("Did not manage to generate a random template that "
                           "satisfies all the constraints")

    def has_atoms_filters(self):
        return len(self.atoms_filters) > 0

    def get_fixed_volume_templates(self, num_prim_cells=10, num_templates=10):
        # Set up a filter that listens to the templates with fixed volume
        from ase.build.tools import niggli_reduce_cell
        cells = []
        transform_matrices = []

        equiv_filter = EquivalentCellsFilter(cells)
        self.add_cell_filter(equiv_filter)

        ucell = self.prim_cell.get_cell()
        matrices = list(all_integer_transform_matrices(num_prim_cells))
        shuffle(matrices)
        for mat in matrices:
            sc = mat.dot(ucell)
            sc, _ = niggli_reduce_cell(sc)
            if self.is_valid(cell=sc):
                # If Atoms filters are present we check if it is valid
                at_valid = True
                if self.has_atoms_filters():
                    atoms = make_supercell(self.prim_cell, mat)
                    at_valid = self.is_valid(atoms=atoms)
                if at_valid:
                    cells.append(sc)
                    transform_matrices.append(mat)

            if len(transform_matrices) >= num_templates:
                break

        templates = []

        if len(transform_matrices) > num_templates:
            shuffle(transform_matrices)
            transform_matrices = transform_matrices[:num_templates]

        for P in transform_matrices:
            atoms = make_supercell(self.prim_cell, P)
            templates.append(atoms)

        # Remove the filter that was artificially added
        self.cell_filters.remove(equiv_filter)
        return templates


def is_3x3_matrix(array):
    return np.array(array).shape == (3, 3)


def check_valid_conversion_matrix(array):
    """
    Make sure that we have a right-handed coordinate system.
    Raise a ValueError if the matrix is not valid.
    """
    determinant = np.linalg.det(array)
    if determinant < 0.0:
        raise ValueError(f"The determinant of the size matrix is less than "
                         f"zero (got {determinant}). For a right coordinate "
                         f"system, we need a positive determinant.")



"""Deprecated import of structure_generator.
Module has been moved to clease.structgen, so
use clease.structgen.structure_generator instead"""

from warnings import warn
from deprecated import deprecated
import numpy as np
from clease.structgen import structure_generator

MSG = 'Import {} from clease.structgen.structure_generator instead'
DEP_VERSION = '0.10.2'  # Deprecation version

MODULE_DEP_MSG = ("""
The clease.new_struct module has been moved as of version {}.
Please use the clease.structgen.structure_generator module instead""".format(DEP_VERSION))

# Print the message on import
warn(MODULE_DEP_MSG, np.VisibleDeprecationWarning)

# pylint: disable=too-few-public-methods


@deprecated(version=DEP_VERSION, reason=MSG.format('StructureGenerator'))
class StructureGenerator(structure_generator.StructureGenerator):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('GSStructure'))
class GSStructure(structure_generator.GSStructure):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('MetropolisTrajectory'))
class MetropolisTrajectory(structure_generator.MetropolisTrajectory):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('ProbeStructure'))
class ProbeStructure(structure_generator.ProbeStructure):
    pass



from clease.calculator import attach_calculator
from clease.montecarlo import Montecarlo
from clease.montecarlo.constraints import ConstrainSwapByBasis
from clease.montecarlo.observers import EnergyEvolution
from clease.montecarlo.observers import EnergyPlotUpdater
from clease.tools import wrap_and_sort_by_position
from clease.gui.constants import SYSTEMS_FROM_DB
import traceback
from ase.db import connect


class MCRunner(object):
    """
    Object for running MC on separate thread

    Parameters:
    atoms: Atoms
        Atoms object to use for MC sampling

    eci: dict
        Dictionary with the effective cluster interactions

    conc: list
        List with concentrations per basis. If we have two basis
        with 3 species in one basis and two in the other, an example
        is [(0.5, 0.3, 0.2), (0.2, 0.8)]. It is important that
        the values sum to one in each basis.

    temps: list
        List with temperatures

    settings: ClusterExpansionSettings
        Settings object holder the required information

    sweeps: int
        Number of MC sweeps per temperature

    status: Label
        Status label to output status messages in the GUI

    db_name: str
        Database to be used when storing MC data

    conc_mode: int
        One of SYSTEMS_FROM_DB or CONC_PER_BASIS. In the first case,
        the concentrations are read from an ASE database. In the latter
        the atoms object is initialised randomly such that it has the
        correct concnetration.

    next_mc_obj: MCRunner
        In case of running many MC calculations, MCRunners can be
        chained. If given, the run function of next_mc_obj will be
        launched when this object finished.
    db_id: int
        If conc_mode is SYSTEMS_FROM_DB the concentration is read from
        the item in the database with this ID.
    """

    def __init__(self,
                 atoms=None,
                 eci=None,
                 mc_page=None,
                 conc=None,
                 temps=None,
                 settings=None,
                 sweeps=None,
                 status=None,
                 db_name=None,
                 conc_mode=None,
                 next_mc_obj=None,
                 db_id=None,
                 app_root=None):
        self.atoms = wrap_and_sort_by_position(atoms)
        self.settings = settings
        self.eci = eci
        self.mc_page = mc_page
        self.conc = conc
        self.conc = conc
        self.temps = temps
        self.num_per_basis = None
        self.basis_elements = settings.concentration.basis_elements
        self.sweeps = sweeps
        self.status = status
        self.orig_template = self.settings.atoms.copy()
        self.db_name = db_name
        self.conc_mode = conc_mode
        self.db_id = db_id
        self.next_mc_obj = next_mc_obj
        self.app_root = app_root

    def _attach_calc(self):
        self.status.text = 'Attaching calculator...'

        # Temporarily disable info update during initalisation
        self.app_root.info_update_disabled = True
        self.app_root.view_mc_cell_disabled = True
        self.atoms = attach_calculator(settings=self.settings, atoms=self.atoms, eci=self.eci)
        self.app_root.active_template_is_mc_cell = True
        self.app_root.info_update_disabled = False
        self.app_root.view_mc_cell_disabled = False

    def _init_conc(self):
        if self.conc_mode == SYSTEMS_FROM_DB:
            self._init_conc_from_db()
            return

        ibb = self.settings.index_by_basis
        for conc_per_basis, indices, symbs in zip(self.conc, ibb, self.basis_elements):
            start = 0
            for c, s in zip(conc_per_basis[1:], symbs[1:]):
                num = int(c * len(indices))
                for i in range(start, num + start):
                    self.atoms[indices[i]].symbol = s
                start += num
        formula = self.atoms.get_chemical_formula()
        msg = f"Running MC at fixed conc for {formula}"
        self.status.text = msg

    def _init_conc_from_db(self):
        db = connect(self.db_name)
        atoms = db.get(id=self.db_id).toatoms()
        atoms = wrap_and_sort_by_position(atoms)

        if len(atoms) != len(self.atoms):
            msg = 'The atoms in the database has to match exactly the '
            msg += ''
            raise ValueError("Currently , ")

        for atom in atoms:
            self.atoms[atom.index].symbol = atom.symbol

    def write_thermodynamic_data_to_db(self, thermo):
        if self.db_name is None:
            return

        if self.db_name == '':
            return

        float_thermo = {}
        for k, v in thermo.items():
            try:
                float_v = float(v)
                float_thermo[k] = float_v
            except Exception:
                pass
        db = connect(self.db_name)

        if self.conc_mode == SYSTEMS_FROM_DB:
            db.update(self.db_id, external_tables={'thermo_data': float_thermo})
        else:
            db.write(self.atoms, external_tables={'thermo_data': float_thermo})

    def run(self):
        try:
            self.mc_page.mc_is_running = True

            self._attach_calc()
            self._init_conc()

            mc = Montecarlo(self.atoms, 200)
            energy_evol = EnergyEvolution(mc, ignore_reset=True)

            energy_update_rate = 5 * len(self.atoms)
            mc.attach(energy_evol, interval=energy_update_rate)

            energy_plot = EnergyPlotUpdater(energy_obs=energy_evol,
                                            graph=self.mc_page.energy_graph,
                                            mean_plot=self.mc_page.mean_energy_plot)
            mc.attach(energy_plot, interval=energy_update_rate)

            cnst = ConstrainSwapByBasis(self.atoms, self.settings.index_by_basis)
            mc.add_constraint(cnst)
            self.mc_page.bind_mc(mc)

            for T in self.temps:
                mc.T = T
                self.status.text = f"Current temperature {T} K"
                mc.run(steps=self.sweeps * len(self.atoms))
                thermo = mc.get_thermodynamic_quantities()
                self.write_thermodynamic_data_to_db(thermo)

            # Reset the old template
            self.settings.set_active_template(atoms=self.orig_template)
            self.app_root.active_template_is_mc_cell = False
            self.mc_page.detach_mc()
            self.status.text = 'MC calculation finished'

            if self.next_mc_obj is not None:
                # Start the next MC runner
                self.next_mc_obj.run()
        except Exception as exc:
            traceback.print_exc()
            self.status.text = str(exc)

        self.mc_page.mc_is_running = False
